"""Define constants for the Ambient PWS component."""
DOMAIN = 'ambient_station'

ATTR_LAST_DATA = 'last_data'

CONF_APP_KEY = 'app_key'

DATA_CLIENT = 'data_client'

TOPIC_UPDATE = 'update'

UNITS_SI = 'si'
UNITS_US = 'us'
